from tox_docker.tests.util import find_container


def test_it_can_run_two_of_the_same_image() -> None:
    hc_builtin = find_container("networking-one")
    hc_custom = find_container("networking-two")

    assert hc_builtin.id != hc_custom.id
    assert hc_builtin.attrs["Image"] == hc_custom.attrs["Image"]


def test_container_has_unique_runas_suffix() -> None:
    container = find_container("healthcheck-builtin")
    assert container.name.startswith("healthcheck-builtin")
    assert container.name != "healthcheck-builtin"
    # don't test the details of how we do this; the suffix could be some
    # randomly-chosen suffix, could be PID (it is), or something else



import os

from docker.models.containers import Container
import docker
import pytest

from tox_docker.config import runas_name


def find_container(instance_name: str) -> Container:
    # TODO: refactor this as a pytest fixture

    # this is running in a child-process of the tox instance which
    # spawned the container; so we need to pass the parent pid to
    # get the right runas_name()
    running_name = runas_name(instance_name, pid=os.getppid())
    client = docker.from_env(version="auto")
    for container in client.containers.list():
        container.attrs["Config"].get("Labels", {})
        if container.name == running_name:
            return container

    pytest.fail(f"No running container with instance name {running_name!r}")



import pytest

from tox_docker.config import Link, runas_name
from tox_docker.tests.util import find_container


def test_linked_containers_can_communicate() -> None:
    outer_container = find_container("networking-two")
    # the outer container should have a link named "linked_host"
    # to the inner container, listening on port 1234
    exitcode, _ = outer_container.exec_run("nc linked_host 1234")
    assert exitcode == 0


def test_link_parsing() -> None:
    assert Link("httpd").target == runas_name("httpd")
    assert Link("httpd").alias == "httpd"

    assert Link("httpd:apache").target == runas_name("httpd")
    assert Link("httpd:apache").alias == "apache"


def test_link_parsing_rejects_trailing_colon() -> None:
    with pytest.raises(ValueError):
        Link("httpd:")



from tox_docker.config import ContainerConfig, ExposedPort, Image
from tox_docker.plugin import escape_env_var, get_port_env_var


def test_escape_env_var() -> None:
    assert (
        escape_env_var("my.private.registry/cat/image")
        == "MY_PRIVATE_REGISTRY_CAT_IMAGE"
    )


def test_get_port_env_var() -> None:
    config = ContainerConfig(
        name="foobar",
        image=Image("foobar"),
        dockerfile=None,
        dockerfile_target="",
        stop=True,
        expose=[
            ExposedPort("CUSTOM_ENV_VAR=1234/tcp"),
        ],
    )

    assert "CUSTOM_ENV_VAR" == get_port_env_var(config, "1234/tcp")

    # fall back to the default name if the port isn't specifically mapped
    assert "FOOBAR_5678_TCP_PORT" == get_port_env_var(config, "5678/tcp")



from pathlib import Path
from typing import Collection, Dict, List, Mapping, Optional
import os
import os.path
import re

from docker.models.containers import Container as DockerContainer
from docker.models.images import Image as DockerImage
from docker.types import Mount
from tox.config.sets import ConfigSet

# nanoseconds in a second; named "SECOND" so that "1.5 * SECOND" makes sense
SECOND = 1000000000

RunningContainers = Dict[str, DockerContainer]

IMAGE_NAME = re.compile(
    # adapted from https://stackoverflow.com/a/39672069, used under CC-BY-SA
    r"^"
    r"("
    r"(?:(?=[^:\/]{1,253})(?!-)[a-zA-Z0-9-]{1,63}(?<!-)(?:\.(?!-)[a-zA-Z0-9-]{1,63}(?<!-))*(?::[0-9]{1,5})?/)?"
    r"(?![._-])(?:[a-z0-9._-]*)(?<![._-])(?:/(?![._-])[a-z0-9._-]*(?<![._-]))*"
    r")"
    r"(?::((?![.-])[a-zA-Z0-9_.-]{1,128}))?"
    r"$"
)

ENV_VAR = re.compile("[A-Z0-9_]+")


def runas_name(container_name: str, pid: Optional[int] = None) -> str:
    """
    Generate a name safe for use in parallel scenarios

    This can be either separate invocations of tox (eg on a busy CI server)
    or `tox -p`. The returned name will be stable for a given container name,
    to avoid needing to maintain a mapping.

    """
    pid = pid or os.getpid()
    return f"{container_name}-tox-{pid}"


class Image:
    def __init__(self, config_line: str) -> None:
        match = IMAGE_NAME.match(config_line)
        if not match:
            raise ValueError(f"{config_line!r} is not a valid image name")
        self.name, self.tag = match.groups()

    def __str__(self) -> str:
        if self.tag:
            return f"{self.name}:{self.tag}"
        return self.name

    def __repr__(self) -> str:
        return repr(str(self))


class Dockerfile:
    def __init__(self, config_line: str) -> None:
        self.path = Path(config_line)
        self.directory = str(self.path.parent)
        self.filename = str(self.path.name)

    def __repr__(self) -> str:
        return str(self.path)


class ExposedPort:
    def __init__(self, config_line: str) -> None:
        env_var, _, container_port_proto = config_line.partition("=")
        container_port, _, protocol = container_port_proto.partition("/")

        if not ENV_VAR.match(env_var):
            raise ValueError(f"{env_var} is not a valid environment variable")
        if not container_port.isdigit():
            raise ValueError("container port must be an int")
        if protocol.lower() not in ("tcp", "udp"):
            raise ValueError("protocol must be tcp or udp")

        self.env_var = env_var
        self.container_port = container_port
        self.protocol = protocol

    @property
    def container_port_proto(self) -> str:
        return f"{self.container_port}/{self.protocol}"


class HostVar:
    def __init__(self, config_line: str) -> None:
        if not ENV_VAR.match(config_line):
            raise ValueError(f"{config_line!r} is not a valid environment variable")
        self.host_var = config_line

    def __str__(self) -> str:
        return self.host_var

    def __repr__(self) -> str:
        return repr(str(self))


class ContainerVar:
    def __init__(self, config_line: str) -> None:
        if not ENV_VAR.match(config_line):
            raise ValueError(f"{config_line!r} is not a valid environment variable")
        self.container_var = config_line

    def __str__(self) -> str:
        return self.container_var

    def __repr__(self) -> str:
        return repr(str(self))


class Link:
    def __init__(self, config_line: str) -> None:
        target, sep, alias = config_line.partition(":")

        if sep and not alias:
            raise ValueError(f"Link '{target}:' missing alias")

        self.target = runas_name(target)

        # this is what the target will be known as INSIDE the
        # container, so don't substitute the runas_name here
        self.alias = alias or target


class Volume:
    def __init__(self, config_line: str) -> None:
        parts = config_line.split(":")
        if len(parts) != 4:
            raise ValueError(f"Volume {config_line!r} is malformed")
        if parts[0] != "bind":
            raise ValueError(f"Volume {config_line!r} type must be 'bind:'")
        if parts[1] not in ("ro", "rw"):
            raise ValueError(f"Volume {config_line!r} options must be 'ro' or 'rw'")

        volume_type, mode, outside, inside = parts
        if not os.path.isabs(outside):
            raise ValueError(f"Volume source {outside!r} must be an absolute path")
        elif not os.path.exists(outside) and parts[1] == "rw":
            os.makedirs(outside, exist_ok=True)
        if not os.path.isabs(inside):
            raise ValueError(f"Mount point {inside!r} must be an absolute path")

        self.docker_mount = Mount(
            source=outside,
            target=inside,
            type=volume_type,
            read_only=bool(mode == "ro"),
        )


class ContainerConfig:
    def __init__(
        self,
        name: str,
        image: Optional[Image],
        dockerfile: Optional[Dockerfile],
        dockerfile_target: str,
        stop: bool,
        environment: Optional[Mapping[str, str]] = None,
        healthcheck_cmd: Optional[str] = None,
        healthcheck_interval: Optional[float] = None,
        healthcheck_timeout: Optional[float] = None,
        healthcheck_start_period: Optional[float] = None,
        healthcheck_retries: Optional[int] = None,
        expose: Optional[Collection[ExposedPort]] = None,
        host_var: Optional[HostVar] = None,
        container_var: Optional[ContainerVar] = None,
        links: Optional[Collection[Link]] = None,
        volumes: Optional[Collection[Volume]] = None,
    ) -> None:
        self.name = name
        self.runas_name = runas_name(name)
        self.image = image
        self.dockerfile = dockerfile
        self.dockerfile_target = dockerfile_target
        self.stop = stop
        self.environment: Mapping[str, str] = environment or {}
        self.expose: Collection[ExposedPort] = expose or []
        self.host_var = str(host_var) if host_var else ""
        self.container_var = str(container_var) if container_var else ""
        self.links: Collection[Link] = links or []
        self.mounts: Collection[Mount] = [v.docker_mount for v in volumes or ()]

        self.healthcheck_cmd = healthcheck_cmd
        self.healthcheck_interval = (
            int(healthcheck_interval) if healthcheck_interval else None
        )
        self.healthcheck_timeout = (
            int(healthcheck_timeout) if healthcheck_timeout else None
        )
        self.healthcheck_start_period = (
            int(healthcheck_start_period) if healthcheck_start_period else None
        )
        self.healthcheck_retries = healthcheck_retries

        self.runnable_image: Optional[DockerImage] = None


class MissingRequiredSetting(Exception):
    pass


def image_required(image: Image) -> Image:
    if not image.name:
        raise MissingRequiredSetting("image")

    return image


class DockerConfigSet(ConfigSet):
    def register_config(self) -> None:
        self.add_config(
            keys=["image"],
            of_type=Optional[Image],
            default=None,
            desc="docker image to run [specify one of image or dockerfile]",
        )
        self.add_config(
            keys=["dockerfile"],
            of_type=Optional[Dockerfile],
            default=None,
            desc="Dockerfile to build/run [specify one of image or dockerfile]",
        )
        self.add_config(
            keys=["dockerfile_target"],
            of_type=str,
            default="",
            desc="Dockerfile target to build/run",
        )
        self.add_config(
            keys=["environment"],
            of_type=Dict[str, str],
            default={},
            desc="environment variables to pass to the docker container",
        )
        self.add_config(
            keys=["expose"],
            of_type=List[ExposedPort],
            default=[],
            desc="container ports to expose to the testenv",
        )
        self.add_config(
            keys=["host_var"],
            of_type=Optional[HostVar],
            default=None,
            desc="environment variable to pass hostname or IP of container to testenv",
        )
        self.add_config(
            keys=["container_var"],
            of_type=Optional[ContainerVar],
            default=None,
            desc="environment variable to pass the name of the container to testenv",
        )
        self.add_config(
            keys=["links"],
            of_type=List[Link],
            default=[],
            desc="containers to link",
        )
        self.add_config(
            keys=["volumes"],
            of_type=List[Volume],
            default=[],
            desc="volumes to attach",
        )

        self.add_config(
            keys=["healthcheck_cmd"],
            of_type=str,
            default="",
            desc="docker healthcheck command",
        )
        self.add_config(
            keys=["healthcheck_interval"],
            of_type=float,
            default=0,
            desc="docker healthcheck interval",
            post_process=lambda num: int(num * SECOND),
        )
        self.add_config(
            keys=["healthcheck_timeout"],
            of_type=float,
            default=0,
            desc="docker healthcheck timeout",
            post_process=lambda num: int(num * SECOND),
        )
        self.add_config(
            keys=["healthcheck_start_period"],
            of_type=float,
            default=0,
            desc="docker healthcheck startup grace period",
            post_process=lambda num: int(num * SECOND),
        )
        self.add_config(
            keys=["healthcheck_retries"],
            of_type=int,
            default=0,
            desc="docker healthcheck retry count",
        )


def parse_container_config(docker_config: DockerConfigSet) -> ContainerConfig:
    if docker_config["image"] and docker_config["dockerfile"]:
        raise ValueError(f"{docker_config.name}: specify image or dockerfile, not both")
    elif not docker_config["image"] and not docker_config["dockerfile"]:
        raise ValueError(f"{docker_config.name}: specify one of image or dockerfile")

    if docker_config["dockerfile_target"] and not docker_config["dockerfile"]:
        raise ValueError(
            f"{docker_config.name}: dockerfile_target specified, but no dockerfile"
        )

    return ContainerConfig(
        name=docker_config.name,
        image=docker_config["image"],
        dockerfile=docker_config["dockerfile"],
        dockerfile_target=docker_config["dockerfile_target"],
        stop=docker_config.name not in docker_config._conf.options.docker_dont_stop,
        environment=docker_config["environment"],
        healthcheck_cmd=docker_config["healthcheck_cmd"],
        healthcheck_interval=docker_config["healthcheck_interval"],
        healthcheck_timeout=docker_config["healthcheck_timeout"],
        healthcheck_start_period=docker_config["healthcheck_start_period"],
        healthcheck_retries=docker_config["healthcheck_retries"],
        expose=docker_config["expose"],
        host_var=docker_config["host_var"],
        links=docker_config["links"],
        volumes=docker_config["volumes"],
    )



from setuptools import find_packages, setup

setup(
    name="tox-docker-id",
    description="Manage lifecycle of docker containers during Tox test runs (patched version)",
    long_description=open("README.rst").read(),
    url="https://github.com/d9pouces/tox-docker",
    maintainer="Matthieu Gallet",
    maintainer_email="github@19pouces.net",
    install_requires=[
        "docker>=4.0,<8.0",
        "tox>=4.0.0,<5.0",
    ],
    packages=find_packages(),
    entry_points={"tox": ["docker = tox_docker"]},
    vcversioner={"version_module_paths": ["_version.py"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Plugins",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Software Development :: Testing",
    ],
)



__all__ = (
    "tox_add_env_config",
    "tox_add_option",
    "tox_after_run_commands",
    "tox_before_run_commands",
)

from logging import getLogger
from typing import Dict, Iterable, List, Mapping, Optional, Tuple, Union
import os
import socket
import sys
import time

from docker.errors import ImageNotFound, NotFound
from docker.models.containers import Container
from tox.config.cli.parser import ToxParser
from tox.config.loader.section import Section
from tox.config.sets import EnvConfigSet
from tox.execute.api import Outcome
from tox.plugin import impl
from tox.session.state import State
from tox.tox_env.api import ToxEnv
from tox.tox_env.errors import Fail
import docker as docker_module

from tox_docker.config import (
    ContainerConfig,
    DockerConfigSet,
    parse_container_config,
    RunningContainers,
)


def log(line: str) -> None:
    getLogger().warning(f"docker> {line}")


class HealthCheckFailed(Exception):
    pass


def get_gateway_ip(container: Container) -> str:
    gateway = os.getenv("TOX_DOCKER_GATEWAY")
    if gateway:
        ip = socket.gethostbyname(gateway)
    elif sys.platform == "darwin":
        # https://docs.docker.com/docker-for-mac/networking/#use-cases-and-workarounds:
        # there is no bridge network available in Docker for Mac, and exposed ports are
        # made available on localhost (but 0.0.0.0 works just as well)
        ip = "0.0.0.0"
    else:
        ip = container.attrs["NetworkSettings"]["Gateway"] or "0.0.0.0"
    return ip


def escape_env_var(varname: str) -> str:
    """
    Convert a string to a form suitable for use as an environment variable.

    The result will be all uppercase, and will have all invalid characters
    replaced by an underscore.

    The result will match the following regex: [a-zA-Z_][a-zA-Z0-9_]*

    Example:
        "my.private.registry/cat/image" will become
        "MY_PRIVATE_REGISTRY_CAT_IMAGE"
    """
    varletters = list(varname.upper())
    if not varletters[0].isalpha():
        varletters[0] = "_"
    for i, c in enumerate(varletters):
        if not c.isalnum() and c != "_":
            varletters[i] = "_"
    return "".join(varletters)


def get_host_env_var(container_config: ContainerConfig) -> str:
    if container_config.host_var:
        return container_config.host_var

    return escape_env_var(f"{container_config.name}_HOST")


def get_container_env_var(container_config: ContainerConfig) -> str:
    if container_config.container_var:
        return container_config.container_var

    return escape_env_var(f"{container_config.name}_CONTAINER")


def get_env_vars(
    container_config: ContainerConfig, container: Container
) -> Mapping[str, str]:
    env = {}
    for containerport, hostports in container.attrs["NetworkSettings"]["Ports"].items():
        if hostports is None:
            # The port is exposed by the container, but not published.
            continue

        for spec in hostports:
            if spec["HostIp"] == "0.0.0.0":
                hostport = spec["HostPort"]
                env_var = get_port_env_var(container_config, containerport)
                env[env_var] = hostport
                break

    gateway_ip = get_gateway_ip(container)
    env_var = get_host_env_var(container_config)
    env[env_var] = gateway_ip
    env_var = get_container_env_var(container_config)
    env[env_var] = container_config.runas_name
    return env


def get_port_env_var(container_config: ContainerConfig, containerport: str) -> str:
    for exposed_port in container_config.expose:
        if exposed_port.container_port_proto == containerport:
            return exposed_port.env_var

    return escape_env_var(f"{container_config.name}_{containerport}_PORT")


def docker_build_or_pull(container_config: ContainerConfig) -> None:
    if container_config.image:
        docker_pull(container_config)
    else:
        docker_build(container_config)


def docker_pull(container_config: ContainerConfig) -> None:
    assert container_config.image

    docker = docker_module.from_env(version="auto")

    try:
        docker.images.get(str(container_config.image))
    except ImageNotFound:
        log(f"pull {container_config.image!r} (from {container_config.name!r})")
        docker.images.pull(container_config.image.name, tag=container_config.image.tag)

    container_config.runnable_image = docker.images.get(str(container_config.image))


def docker_build(container_config: ContainerConfig) -> None:
    assert container_config.dockerfile

    docker = docker_module.from_env(version="auto")

    if container_config.dockerfile_target:
        log(
            f"build {container_config.dockerfile!r} target {container_config.dockerfile_target!r}"
        )
    else:
        log(f"build {container_config.dockerfile!r}")

    image, _ = docker.images.build(
        path=container_config.dockerfile.directory,
        dockerfile=container_config.dockerfile.filename,
        target=container_config.dockerfile_target or None,
        pull=True,
        forcerm=True,
    )
    log(f"built: {image.short_id}")

    container_config.runnable_image = image


def docker_run(
    container_config: ContainerConfig,
    running_containers: RunningContainers,
) -> Container:
    docker = docker_module.from_env(version="auto")

    healthcheck: Dict[str, Union[List[str], int]] = {}
    if container_config.healthcheck_cmd:
        healthcheck["test"] = ["CMD-SHELL", container_config.healthcheck_cmd]
    if container_config.healthcheck_interval:
        healthcheck["interval"] = container_config.healthcheck_interval
    if container_config.healthcheck_timeout:
        healthcheck["timeout"] = container_config.healthcheck_timeout
    if container_config.healthcheck_start_period:
        healthcheck["start_period"] = container_config.healthcheck_start_period
    if container_config.healthcheck_retries:
        healthcheck["retries"] = container_config.healthcheck_retries

    ports = {p.container_port_proto: 0 for p in container_config.expose}

    links = {}
    for link in container_config.links:
        if link.target not in running_containers:
            raise ValueError(
                f"Container {link.target!r} not running; it must come before {container_config.name!r} in the docker= list"
            )
        other_container = running_containers[link.target]
        links[other_container.id] = link.alias

    for mount in container_config.mounts:
        source = mount["Source"]
        if not os.path.exists(source):
            raise ValueError(f"Volume source {source!r} does not exist")

    assert container_config.runnable_image
    image_name = container_config.image or container_config.runnable_image.short_id
    log(f"run {image_name!r} (from {container_config.name!r})")

    container = docker.containers.run(
        container_config.runnable_image.id,
        name=container_config.runas_name,
        detach=True,
        environment=container_config.environment,
        healthcheck=healthcheck or None,
        links=links,
        ports=ports,
        publish_all_ports=len(ports) == 0,
        mounts=container_config.mounts,
    )
    container.reload()  # TODO: why do we need this?
    return container


def docker_health_check(
    container_config: ContainerConfig, container: Container
) -> None:
    docker = docker_module.from_env(version="auto")

    if "Health" in container.attrs["State"]:
        log(f"health check {container_config.name!r}")
        while True:
            container.reload()
            health = container.attrs["State"]["Health"]["Status"]
            if health == "healthy":
                break
            elif health == "starting":
                time.sleep(0.1)
            elif health == "unhealthy":
                # the health check failed after its own timeout
                msg = f"{container_config.image!r} (from {container_config.name!r}) failed health check"
                raise HealthCheckFailed(msg)


def docker_stop(container_config: ContainerConfig, container: Container) -> None:
    if container_config.stop:
        log(f"remove '{container.short_id}' (from {container_config.name!r})")
        container.remove(v=True, force=True)
    else:
        log(f"leave '{container.short_id}' (from {container_config.name!r}) running")


def docker_get(container_config: ContainerConfig) -> Optional[Container]:
    docker = docker_module.from_env(version="auto")
    try:
        return docker.containers.get(container_config.runas_name)
    except NotFound:
        return None


def stop_containers(containers: Iterable[Tuple[ContainerConfig, Container]]) -> None:
    for container_config, container in containers:
        docker_stop(container_config, container)


@impl
def tox_add_env_config(env_conf: EnvConfigSet, state: State) -> None:
    def build_docker_config_set(container_name: object) -> DockerConfigSet:
        assert isinstance(container_name, str)
        docker_conf = state.conf.get_section_config(
            section=Section("docker", container_name),
            base=[],
            of_type=DockerConfigSet,
            for_env=None,
        )
        if not docker_conf.loaders:
            raise ValueError(f"Missing [docker:{container_name}] in tox.ini")
        return docker_conf

    env_conf.add_config(
        keys=["docker"],
        of_type=List[DockerConfigSet],
        default=[],
        desc="docker image configs to load",
        factory=build_docker_config_set,  # type: ignore
    )


@impl
def tox_before_run_commands(tox_env: ToxEnv) -> None:
    docker_confs = tox_env.conf.load("docker")

    container_configs = [
        parse_container_config(docker_conf) for docker_conf in docker_confs
    ]

    seen = set()
    for container_config in container_configs:
        if container_config.name in seen:
            raise ValueError(
                f"Container {container_config.name!r} specified more than once"
            )
        seen.add(container_config.name)

    for container_config in container_configs:
        docker_build_or_pull(container_config)

    config_and_container = []
    running_containers: RunningContainers = {}
    for container_config in container_configs:
        container = docker_run(container_config, running_containers)
        config_and_container.append((container_config, container))
        running_containers[container_config.runas_name] = container

    for container_config, container in config_and_container:
        try:
            docker_health_check(container_config, container)
        except HealthCheckFailed:
            tox_env.interrupt()
            clean_up_containers(tox_env)
            raise Fail(
                f"{container_config.image!r} (from {container_config.name!r}) failed health check"
            )

        tox_env.conf["set_env"].update(get_env_vars(container_config, container))


@impl
def tox_after_run_commands(
    tox_env: ToxEnv, exit_code: int, outcomes: List[Outcome]
) -> None:
    clean_up_containers(tox_env)


def clean_up_containers(tox_env: ToxEnv) -> None:
    docker_confs = tox_env.conf.load("docker")

    container_configs = [
        parse_container_config(docker_conf) for docker_conf in docker_confs
    ]

    configs_and_containers = []
    for config in container_configs:
        container = docker_get(config)
        if container:
            configs_and_containers.append((config, container))

    stop_containers(configs_and_containers)


@impl
def tox_add_option(parser: ToxParser) -> None:
    # command line flag to keep docker containers running
    parser.add_argument(
        "--docker-dont-stop",
        default=[],
        action="append",
        metavar="CONTAINER",
        help=(
            "If specified, tox-docker will not stop CONTAINER after the test run. "
            "Can be specified multiple times."
        ),
    )



from urllib.request import urlopen
import os


def test_default_host_name_var() -> None:
    # [docker:healthcheck-builtin] doesn't specify host_var
    assert "HEALTHCHECK_BUILTIN_HOST" in os.environ


def test_custom_host_name_var() -> None:
    # [docker:healthcheck-custom] _does_ specify host_var=...
    assert "NET_TWO_CUSTOM_HOST" in os.environ


def test_EXPOSEd_ports_are_available_when_expose_is_not_set() -> None:
    # [docker:healthcheck-builtin] does not have expose=
    assert "HEALTHCHECK_BUILTIN_8000_TCP_PORT" in os.environ


def test_EXPOSEd_ports_are_not_available_if_not_listed_in_expose() -> None:
    # [docker:networking-two] doesn't map 5678/udp, but it is EXPOSEd
    assert "NETWORKING_TWO_5678_UDP_PORT" not in os.environ


def test_manually_mapped_ports_dont_get_default_port_envvar() -> None:
    # [docker:networking-two] explicitly maps 1234/tcp
    assert "NETWORKING_TWO_1234_TCP_PORT" not in os.environ


def test_exposed_ports_are_accessible_to_test_runs() -> None:
    host = os.environ["HEALTHCHECK_BUILTIN_HOST"]
    port = os.environ["HEALTHCHECK_BUILTIN_8000_TCP_PORT"]

    response = urlopen(f"http://{host}:{port}/")
    assert response.getcode() == 200
    assert b"Directory listing for /" in response.read()
